@@ -1,40 +1,131 @@
-use heck::ToUpperCamelCase;
+use std::collections::HashMap;
 
 use error::Error;
+use heck::ToUpperCamelCase;
 
 use crate::{
-    parse::{Quantifier, TypedParam},
-    pg_type,
+    parse::Quantifier,
+    pg_type::{CornucopiaType, CornucopiaTypeKind, TypeRegistrar},
     prepare_queries::{PreparedQuery, RustReturnType},
 };
 
 use super::prepare_queries::PreparedModule;
 
-pub fn generate_query(query: &PreparedQuery) -> Result<String, Error> {
+pub(crate) fn generate_query(
+    module_name: &str,
+    query: &PreparedQuery,
+) -> Result<(String, String), Error> {
     let name = &query.name;
     let query_struct = generate_query_struct(query)?.unwrap_or_default();
     let params = generate_query_params(query)?;
     let ret_ty = generate_query_ret_ty(query)?;
-    let ret = generate_query_quantified_ret_ty(query, &ret_ty);
+    let ret_ty = if let RustReturnType::Struct(_) = query.ret {
+        format!("super::super::queries::{module_name}::{ret_ty}")
+    } else {
+        ret_ty
+    };
 
+    let ret = generate_query_quantified_ret_ty(query, &ret_ty);
     let body = generate_query_body(query, ret_ty)?;
-    Ok(format!(
-        r#"{query_struct}
-pub async fn {name}(client:&Client, {params}) -> Result<{ret},Error> {{{body}}}
-pub async fn {name}_tx<'a>(client:&Transaction<'a>, {params}) -> Result<{ret}, Error> {{{body}}}
-"#
-    ))
+    let query_string = format!(
+        "{query_struct}
+    pub async fn {name}(client:&Client, {params}) -> Result<{ret},Error> {{{body}}}"
+    );
+
+    let transaction_string = format!("pub async fn {name}<'a>(client:&Transaction<'a>, {params}) -> Result<{ret}, Error> {{{body}}}");
+
+    Ok((query_string, transaction_string))
 }
 
-pub fn generate_query_struct(query: &PreparedQuery) -> Result<Option<String>, Error> {
+pub(crate) fn generate_custom_type(ty: &CornucopiaType) -> String {
+    let type_def = match &ty.kind {
+        CornucopiaTypeKind::Enum(variants) => {
+            let name = &ty.rust_ty_name;
+            format!(
+                "#[derive(Clone, Copy, PartialEq, Eq)]\npub enum {} {{ {} }}",
+                name,
+                variants.join(",")
+            )
+        }
+        CornucopiaTypeKind::Domain(domain_inner_ty) => {
+            let name = &domain_inner_ty.rust_ty_name;
+            format!(
+                "#[derive(Clone)]\npub struct {} ({})",
+                ty.pg_ty.schema(),
+                name
+            )
+        }
+        CornucopiaTypeKind::Composite(fields) => {
+            let fields_str = fields
+                .iter()
+                .map(|f| format!("pub {} : {}", f.name, f.ty.rust_ty_definition_path))
+                .collect::<Vec<String>>()
+                .join(",");
+
+            let name = &ty.rust_ty_name;
+            format!("#[derive(Clone)]\npub struct {} {{ {} }}", name, fields_str)
+        }
+        CornucopiaTypeKind::Base => {
+            panic!("got base type while generating custom types. This is a bug")
+        }
+    };
+
+    format!(
+        "#[derive(Debug, ToSql, FromSql)]\n#[postgres(name = \"{}\")]\n{}",
+        ty.pg_ty.name(),
+        type_def
+    )
+}
+
+pub(crate) fn generate_type_modules(type_registrar: &TypeRegistrar) -> String {
+    let mut modules = HashMap::<String, Vec<CornucopiaType>>::new();
+    for ((schema, _), ty) in &type_registrar.custom_types {
+        match modules.entry(schema.to_owned()) {
+            std::collections::hash_map::Entry::Occupied(mut entry) => {
+                entry.get_mut().push(ty.clone());
+            }
+            std::collections::hash_map::Entry::Vacant(entry) => {
+                entry.insert(vec![ty.clone()]);
+            }
+        }
+    }
+    let schema_modules = modules
+        .iter()
+        .map(|(mod_name, tys)| {
+            let tys_str = tys
+                .iter()
+                .map(generate_custom_type)
+                .collect::<Vec<String>>()
+                .join("\n\n");
+            format!(
+                "
+        pub mod {mod_name} {{
+            use postgres_types::{{FromSql, ToSql}};
+
+            {tys_str} 
+        }}"
+            )
+        })
+        .collect::<Vec<String>>()
+        .join("\n\n");
+
+    format!(
+        "pub mod types {{
+    {schema_modules}        
+}}"
+    )
+}
+
+pub(crate) fn generate_query_struct(query: &PreparedQuery) -> Result<Option<String>, Error> {
     if let RustReturnType::Struct(fields) = &query.ret {
         let mut field_strings = Vec::new();
-        for TypedParam { name, ty } in fields {
-            field_strings.push(format!(
-                r#"pub {} : {}"#,
-                name,
-                pg_type::to_equivalent_rust_string(ty)?
-            ));
+        for (param, param_ty) in fields {
+            let ty_string = if param.is_nullable {
+                format!("Option<{}>", param_ty.rust_ty_usage_path)
+            } else {
+                param_ty.rust_ty_usage_path.clone()
+            };
+            field_strings.push(format!(r#"pub {} : {}"#, param.name, ty_string));
         }
         let fields_string = field_strings.join(",");
         let struct_name = query.name.to_upper_camel_case();
@@ -48,7 +139,7 @@ pub struct {struct_name} {{{fields_string}}}"#
     }
 }
 
-pub fn generate_query_quantified_ret_ty(query: &PreparedQuery, ret_ty: &str) -> String {
+pub(crate) fn generate_query_quantified_ret_ty(query: &PreparedQuery, ret_ty: &str) -> String {
     if let RustReturnType::Void = query.ret {
         String::from("()")
     } else {
@@ -60,47 +151,33 @@ pub fn generate_query_quantified_ret_ty(query: &PreparedQuery, ret_ty: &str) ->
     }
 }
 
-pub fn generate_query_params(query: &PreparedQuery) -> Result<String, Error> {
+pub(crate) fn generate_query_params(query: &PreparedQuery) -> Result<String, Error> {
     let mut param_strings = Vec::new();
     for param in &query.params {
-        let rs_ty = pg_type::to_equivalent_rust_string(&param.ty)?;
-        let rs_ty_borrowed = if &pg_type::to_equivalent_rust_string(&param.ty)? == "String" {
-            String::from("&str")
-        } else {
-            format!("&{}", rs_ty)
-        };
-
+        let rs_ty_borrowed = &param.ty.borrowed_rust_ty();
         let param_string = format!("{} : {}", param.name, rs_ty_borrowed);
         param_strings.push(param_string);
     }
     Ok(param_strings.join(","))
 }
 
-pub fn generate_query_ret_ty(query: &PreparedQuery) -> Result<String, Error> {
+pub(crate) fn generate_query_ret_ty(query: &PreparedQuery) -> Result<String, Error> {
     Ok(match &query.ret {
-        crate::prepare_queries::RustReturnType::Void => String::from("()"),
-        crate::prepare_queries::RustReturnType::Scalar(ty) => {
-            pg_type::to_equivalent_rust_string(ty)?
-        }
-        crate::prepare_queries::RustReturnType::Tuple(tys) => {
+        RustReturnType::Void => String::from("()"),
+        RustReturnType::Scalar(ty) => ty.rust_ty_usage_path.clone(),
+        RustReturnType::Tuple(tys) => {
             let mut tys_string = Vec::new();
             for ty in tys {
-                tys_string.push(pg_type::to_equivalent_rust_string(ty)?)
+                tys_string.push(ty.rust_ty_usage_path.clone())
             }
             format!("({})", tys_string.join(","))
         }
-        crate::prepare_queries::RustReturnType::Struct(_) => query.name.to_upper_camel_case(),
+        RustReturnType::Struct(_) => query.name.to_upper_camel_case(),
     })
 }
 
-pub fn generate_query_body(query: &PreparedQuery, ret_ty: String) -> Result<String, Error> {
+pub(crate) fn generate_query_body(query: &PreparedQuery, ret_ty: String) -> Result<String, Error> {
     let query_string = format!(r#""{}""#, &query.sql);
-    let override_types = query
-        .override_types
-        .iter()
-        .map(pg_type::to_litteral_rust_string)
-        .collect::<Vec<String>>()
-        .join(",");
 
     let query_method = if let RustReturnType::Void = query.ret {
         "execute"
@@ -136,21 +213,18 @@ pub fn generate_query_body(query: &PreparedQuery, ret_ty: String) -> Result<Stri
             let mut rust_ret_values = Vec::new();
             let mut rust_ret_value_names = Vec::new();
             for (i, ty) in tup.iter().enumerate() {
-                let rust_ty = pg_type::to_equivalent_rust_string(ty)?;
                 let rust_ret_value_name = format!("return_value_{}", i);
                 rust_ret_values.push(format!(
                     "let {}: {} = res.get({});",
-                    rust_ret_value_name, rust_ty, i
+                    rust_ret_value_name, ty.rust_ty_usage_path, i
                 ));
                 rust_ret_value_names.push(rust_ret_value_name);
             }
-
             let tuple_value_string = format!(
                 "{} ({})",
                 rust_ret_values.join(" "),
                 rust_ret_value_names.join(",")
             );
-
             match query.quantifier {
                 Quantifier::ZeroOrMore => {
                     format!("let return_value = res.iter().map(|res| {{ {tuple_value_string} }}).collect::<Vec<{ret_ty}>>(); Ok(return_value)")
@@ -166,14 +240,18 @@ pub fn generate_query_body(query: &PreparedQuery, ret_ty: String) -> Result<Stri
         RustReturnType::Struct(structure) => {
             let mut field_values = Vec::new();
             let mut rust_ret_values = Vec::new();
-            for (i, field) in structure.iter().enumerate() {
-                let rust_ty = pg_type::to_equivalent_rust_string(&field.ty)?;
+            for (i, (param, param_ty)) in structure.iter().enumerate() {
+                let ty_string = if param.is_nullable {
+                    format!("Option<{}>", param_ty.rust_ty_usage_path)
+                } else {
+                    param_ty.rust_ty_usage_path.clone()
+                };
                 let rust_ret_value_name = format!("return_value_{}", i);
                 rust_ret_values.push(format!(
                     "let {}: {} = res.get({});",
-                    rust_ret_value_name, rust_ty, i
+                    rust_ret_value_name, ty_string, i
                 ));
-                field_values.push(format!("{}: {}", field.name, rust_ret_value_name));
+                field_values.push(format!("{} : {}", param.name, rust_ret_value_name));
             }
 
             let struct_value_string = format!(
@@ -197,41 +275,73 @@ pub fn generate_query_body(query: &PreparedQuery, ret_ty: String) -> Result<Stri
         }
     };
 
+    let res_var_name = match query.ret {
+        RustReturnType::Void => "_",
+        _ => "res",
+    };
+
     Ok(format!(
-        "let stmt = client.prepare_typed_cached({query_string}, &[{override_types}]).await?;
-let res = client.{query_method}(&stmt, &[{query_param_values}]).await?;
+        "let stmt = client.prepare_cached({query_string}).await?;
+let {res_var_name} = client.{query_method}(&stmt, &[{query_param_values}]).await?;
 
 {ret_value}"
     ))
 }
 
-pub fn generate(modules: Vec<PreparedModule>, destination: &str) -> Result<(), Error> {
-    let imports = r#"use deadpool_postgres::{Client, Transaction};
-use tokio_postgres::{types::Type, error::Error};"#;
-    let mut generated_modules = Vec::new();
+pub(crate) fn generate(
+    type_registrar: &TypeRegistrar,
+    modules: Vec<PreparedModule>,
+    destination: &str,
+) -> Result<(), Error> {
+    let query_imports = r#"use deadpool_postgres::Client;
+use tokio_postgres::{error::Error};"#;
+    let transaction_imports = r#"use deadpool_postgres::Transaction;
+    use tokio_postgres::{error::Error};"#;
+
+    let type_modules = generate_type_modules(type_registrar);
+
+    let mut query_modules = Vec::new();
+    let mut transaction_modules = Vec::new();
     for module in modules {
         let mut query_strings = Vec::new();
+        let mut transaction_strings = Vec::new();
         for query in module.queries {
-            query_strings.push(generate_query(&query)?)
+            let (query_string, transaction_string) = generate_query(&module.name, &query)?;
+            query_strings.push(query_string);
+            transaction_strings.push(transaction_string)
         }
         let queries_string = query_strings.join("\n\n");
+        let transactions_string = transaction_strings.join("\n\n");
         let module_name = module.name;
 
-        generated_modules.push(format!(
+        query_modules.push(format!(
             r#"pub mod {module_name} {{
-{imports}
+{query_imports}
 
 {queries_string}
+}}"#
+        ));
+        transaction_modules.push(format!(
+            r#"pub mod {module_name} {{
+{transaction_imports}
+
+{transactions_string}
 }}"#
         ));
     }
 
-    std::fs::write(destination, generated_modules.join("\n\n"))?;
+    let generated_modules = format!(
+        "{type_modules} \n pub mod queries {{ {} }} \n pub mod transactions {{ {} }}",
+        query_modules.join("\n\n"),
+        transaction_modules.join("\n\n")
+    );
+
+    std::fs::write(destination, generated_modules)?;
 
     Ok(())
 }
 
-pub mod error {
+pub(crate) mod error {
     use thiserror::Error as ThisError;
 
     use crate::pg_type::error::UnsupportedPostgresTypeError;
